@@ -12,10 +12,12 @@
 
 #include "include/JSArrayProxy.hh"
 
+#include "include/JSArrayIterProxy.hh"
+
 #include "include/modules/pythonmonkey/pythonmonkey.hh"
 #include "include/jsTypeFactory.hh"
 #include "include/pyTypeFactory.hh"
-#include "include/PyProxyHandler.hh"
+#include "include/PyBaseProxyHandler.hh"
 
 #include <jsapi.h>
 #include <jsfriendapi.h>
@@ -25,7 +27,7 @@
 
 void JSArrayProxyMethodDefinitions::JSArrayProxy_dealloc(JSArrayProxy *self)
 {
-  self->jsObject.set(nullptr);
+  self->jsArray.set(nullptr);
   Py_TYPE(self)->tp_free((PyObject *)self);
 }
 
@@ -43,7 +45,7 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_init(JSArrayProxy *self, PyObjec
 Py_ssize_t JSArrayProxyMethodDefinitions::JSArrayProxy_length(JSArrayProxy *self)
 {
   uint32_t length;
-  JS::GetArrayLength(GLOBAL_CX, self->jsObject, &length);
+  JS::GetArrayLength(GLOBAL_CX, self->jsArray, &length);
   return (Py_ssize_t)length;
 }
 
@@ -55,14 +57,13 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_get(JSArrayProxy *self, Py
     return NULL; // key is not a str or int
   }
 
-  JS::RootedValue *value = new JS::RootedValue(GLOBAL_CX);
-  JS_GetPropertyById(GLOBAL_CX, self->jsObject, id, value);
-
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject));
   // look through the methods for dispatch and return key if no method found
   for (size_t index = 0;; index++) {
     const char *methodName = JSArrayProxyType.tp_methods[index].ml_name;
     if (methodName == NULL) {   // reached end of list
+      JS::RootedValue *value = new JS::RootedValue(GLOBAL_CX);
+      JS_GetPropertyById(GLOBAL_CX, self->jsArray, id, value);
+      JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
       return pyTypeFactory(GLOBAL_CX, global, value)->getPyObject();
     }
     else if (PyUnicode_Check(key)) {
@@ -71,6 +72,9 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_get(JSArrayProxy *self, Py
       }
     }
     else {
+      JS::RootedValue *value = new JS::RootedValue(GLOBAL_CX);
+      JS_GetPropertyById(GLOBAL_CX, self->jsArray, id, value);
+      JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
       return pyTypeFactory(GLOBAL_CX, global, value)->getPyObject();
     }
   }
@@ -83,11 +87,11 @@ static PyObject *list_slice(JSArrayProxy *self, Py_ssize_t ilow, Py_ssize_t ihig
   jArgs[0].setInt32(ilow);
   jArgs[1].setInt32(ihigh);
   JS::RootedValue *jReturnedArray = new JS::RootedValue(GLOBAL_CX);
-  if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "slice", jArgs, jReturnedArray)) {
+  if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "slice", jArgs, jReturnedArray)) {
     PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
     return NULL;
   }
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject)), jReturnedArray)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), jReturnedArray)->getPyObject();
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_get_subscript(JSArrayProxy *self, PyObject *key)
@@ -113,9 +117,9 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_get_subscript(JSArrayProxy
     JS_IndexToId(GLOBAL_CX, index, &id);
 
     JS::RootedValue *value = new JS::RootedValue(GLOBAL_CX);
-    JS_GetPropertyById(GLOBAL_CX, self->jsObject, id, value);
+    JS_GetPropertyById(GLOBAL_CX, self->jsArray, id, value);
 
-    return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject)), value)->getPyObject();
+    return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), value)->getPyObject();
   }
   else if (PySlice_Check(key)) {
     Py_ssize_t start, stop, step, slicelength, index;
@@ -137,14 +141,14 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_get_subscript(JSArrayProxy
 
       JS::RootedValue elementVal(GLOBAL_CX);
       for (size_t cur = start, index = 0; index < slicelength; cur += (size_t)step, index++) {
-        JS_GetElement(GLOBAL_CX, self->jsObject, cur, &elementVal);
+        JS_GetElement(GLOBAL_CX, self->jsArray, cur, &elementVal);
         JS_SetElement(GLOBAL_CX, jCombinedArray, index, elementVal);
       }
 
       JS::RootedValue *jCombinedArrayValue = new JS::RootedValue(GLOBAL_CX);
       jCombinedArrayValue->setObjectOrNull(jCombinedArray);
 
-      return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject)), jCombinedArrayValue)->getPyObject();
+      return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), jCombinedArrayValue)->getPyObject();
     }
   }
   else {
@@ -226,30 +230,30 @@ static int list_ass_slice(JSArrayProxy *self, Py_ssize_t ilow, Py_ssize_t ihigh,
   if (d < 0) {   /* Delete -d items */
     JS::RootedValue elementVal(GLOBAL_CX);
     for (size_t index = ihigh, count = 0; count < selfLength - ihigh; index++, count++) {
-      JS_GetElement(GLOBAL_CX, self->jsObject, index, &elementVal);
-      JS_SetElement(GLOBAL_CX, self->jsObject, index+d, elementVal);
+      JS_GetElement(GLOBAL_CX, self->jsArray, index, &elementVal);
+      JS_SetElement(GLOBAL_CX, self->jsArray, index+d, elementVal);
     }
 
-    JS::SetArrayLength(GLOBAL_CX, self->jsObject, selfLength + d);
+    JS::SetArrayLength(GLOBAL_CX, self->jsArray, selfLength + d);
   }
   else if (d > 0) { /* Insert d items */
     k = selfLength;
 
-    JS::SetArrayLength(GLOBAL_CX, self->jsObject, k + d);
+    JS::SetArrayLength(GLOBAL_CX, self->jsArray, k + d);
 
     selfLength = k + d;
 
     JS::RootedValue elementVal(GLOBAL_CX);
     for (size_t index = ihigh, count = 0; count < k - ihigh; index++, count++) {
-      JS_GetElement(GLOBAL_CX, self->jsObject, index, &elementVal);
-      JS_SetElement(GLOBAL_CX, self->jsObject, index+d, elementVal);
+      JS_GetElement(GLOBAL_CX, self->jsArray, index, &elementVal);
+      JS_SetElement(GLOBAL_CX, self->jsArray, index+d, elementVal);
     }
   }
 
   JS::RootedValue elementVal(GLOBAL_CX);
   for (k = 0; k < n; k++, ilow++) {
     elementVal.set(jsTypeFactory(GLOBAL_CX, vitem[k]));
-    JS_SetElement(GLOBAL_CX, self->jsObject, ilow, elementVal);
+    JS_SetElement(GLOBAL_CX, self->jsArray, ilow, elementVal);
   }
 
   result = 0;
@@ -282,10 +286,10 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_assign_key(JSArrayProxy *self, P
 
     if (value) { // we are setting a value
       JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, value));
-      JS_SetPropertyById(GLOBAL_CX, self->jsObject, id, jValue);
+      JS_SetPropertyById(GLOBAL_CX, self->jsArray, id, jValue);
     } else { // we are deleting a value
       JS::ObjectOpResult ignoredResult;
-      JS_DeletePropertyById(GLOBAL_CX, self->jsObject, id, ignoredResult);
+      JS_DeletePropertyById(GLOBAL_CX, self->jsArray, id, ignoredResult);
     }
 
     return 0;
@@ -341,8 +345,8 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_assign_key(JSArrayProxy *self, P
         }
 
         for (size_t index = cur, count = 0; count < lim; index++, count++) {
-          JS_GetElement(GLOBAL_CX, self->jsObject, index + 1, &elementVal);
-          JS_SetElement(GLOBAL_CX, self->jsObject, index - i, elementVal);
+          JS_GetElement(GLOBAL_CX, self->jsArray, index + 1, &elementVal);
+          JS_SetElement(GLOBAL_CX, self->jsArray, index - i, elementVal);
         }
       }
 
@@ -350,12 +354,12 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_assign_key(JSArrayProxy *self, P
 
       if (cur < (size_t)selfSize) {
         for (size_t index = cur, count = 0; count < selfSize - cur; index++, count++) {
-          JS_GetElement(GLOBAL_CX, self->jsObject, index, &elementVal);
-          JS_SetElement(GLOBAL_CX, self->jsObject, index - slicelength, elementVal);
+          JS_GetElement(GLOBAL_CX, self->jsArray, index, &elementVal);
+          JS_SetElement(GLOBAL_CX, self->jsArray, index - slicelength, elementVal);
         }
       }
 
-      JS::SetArrayLength(GLOBAL_CX, self->jsObject, selfSize - slicelength);
+      JS::SetArrayLength(GLOBAL_CX, self->jsArray, selfSize - slicelength);
 
       return 0;
     }
@@ -395,7 +399,7 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_assign_key(JSArrayProxy *self, P
       JS::RootedValue elementVal(GLOBAL_CX);
       for (cur = start, i = 0; i < slicelength; cur += (size_t)step, i++) {
         elementVal.set(jsTypeFactory(GLOBAL_CX, seqitems[i]));
-        JS_SetElement(GLOBAL_CX, self->jsObject, cur, elementVal);
+        JS_SetElement(GLOBAL_CX, self->jsArray, cur, elementVal);
       }
 
       Py_DECREF(seq);
@@ -444,18 +448,18 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_richcompare(JSArrayProxy *
   }
 
   JS::RootedValue *elementVal;
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject));
+  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
 
   Py_ssize_t index;
   /* Search for the first index where items are different */
   for (index = 0; index < selfLength && index < otherLength; index++) {
     elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsObject, index, elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
 
     PyObject *leftItem = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
     PyObject *rightItem;
     if (PyObject_TypeCheck(other, &JSArrayProxyType)) {
-      JS_GetElement(GLOBAL_CX, ((JSArrayProxy *)other)->jsObject, index, elementVal);
+      JS_GetElement(GLOBAL_CX, ((JSArrayProxy *)other)->jsArray, index, elementVal);
       rightItem = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
     } else {
       rightItem = ((PyListObject *)other)->ob_item[index];
@@ -492,7 +496,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_richcompare(JSArrayProxy *
   }
 
   elementVal = new JS::RootedValue(GLOBAL_CX);
-  JS_GetElement(GLOBAL_CX, self->jsObject, index, elementVal);
+  JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
   /* Compare the final item again using the proper operator */
   return PyObject_RichCompare(pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject(), ((PyListObject *)other)->ob_item[index], op);
 }
@@ -516,7 +520,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repr(JSArrayProxy *self) {
   /* "[" + "1" + ", 2" * (len - 1) + "]" */
   writer.min_length = 1 + 1 + (2 + 1) * (selfLength - 1) + 1;
 
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject));
+  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
 
   if (_PyUnicodeWriter_WriteChar(&writer, '[') < 0) {
     goto error;
@@ -531,10 +535,10 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repr(JSArrayProxy *self) {
     }
 
     JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsObject, index, elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
 
     PyObject *s;
-    if (&elementVal->toObject() == self->jsObject.get()) {
+    if (&elementVal->toObject() == self->jsArray.get()) {
       s = PyObject_Repr((PyObject *)self);
     } else {
       s = PyObject_Repr(pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject());
@@ -565,19 +569,29 @@ error:
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_iter(JSArrayProxy *self) {
-  const Py_ssize_t selfLength = JSArrayProxy_length(self);
-
-  PyObject *seq = PyList_New(selfLength);
-
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject));
-  for (size_t index = 0; index < selfLength; index++) {
-    JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsObject, index, elementVal);
-    PyList_SET_ITEM(seq, index, pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject());
+  JSArrayIterProxy *iterator = PyObject_GC_New(JSArrayIterProxy, &JSArrayIterProxyType);
+  if (iterator == NULL) {
+    return NULL;
   }
+  iterator->it.reversed = false;
+  iterator->it.it_index = 0;
+  Py_INCREF(self);
+  iterator->it.it_seq = (PyListObject *)self;
+  PyObject_GC_Track(iterator);
+  return (PyObject *)iterator;
+}
 
-  // Convert to a Python iterator
-  return PyObject_GetIter(seq);
+PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_iter_reverse(JSArrayProxy *self) {
+  JSArrayIterProxy *iterator = PyObject_GC_New(JSArrayIterProxy, &JSArrayIterProxyType);
+  if (iterator == NULL) {
+    return NULL;
+  }
+  iterator->it.reversed = true;
+  iterator->it.it_index = JSArrayProxyMethodDefinitions::JSArrayProxy_length(self) - 1;
+  Py_INCREF(self);
+  iterator->it.it_seq = (PyListObject *)self;
+  PyObject_GC_Track(iterator);
+  return (PyObject *)iterator;
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_concat(JSArrayProxy *self, PyObject *value) {
@@ -612,13 +626,13 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_concat(JSArrayProxy *self,
   JS::RootedValue elementVal(GLOBAL_CX);
 
   for (Py_ssize_t inputIdx = 0; inputIdx < sizeSelf; inputIdx++) {
-    JS_GetElement(GLOBAL_CX, self->jsObject, inputIdx, &elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, inputIdx, &elementVal);
     JS_SetElement(GLOBAL_CX, jCombinedArray, inputIdx, elementVal);
   }
 
   if (PyObject_TypeCheck(value, &JSArrayProxyType)) {
     for (Py_ssize_t inputIdx = 0; inputIdx < sizeValue; inputIdx++) {
-      JS_GetElement(GLOBAL_CX, ((JSArrayProxy *)value)->jsObject, inputIdx, &elementVal);
+      JS_GetElement(GLOBAL_CX, ((JSArrayProxy *)value)->jsArray, inputIdx, &elementVal);
       JS_SetElement(GLOBAL_CX, jCombinedArray, sizeSelf + inputIdx, elementVal);
     }
   } else {
@@ -632,7 +646,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_concat(JSArrayProxy *self,
   JS::RootedValue *jCombinedArrayValue = new JS::RootedValue(GLOBAL_CX);
   jCombinedArrayValue->setObjectOrNull(jCombinedArray);
 
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject)), jCombinedArrayValue)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), jCombinedArrayValue)->getPyObject();
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repeat(JSArrayProxy *self, Py_ssize_t n) {
@@ -650,7 +664,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repeat(JSArrayProxy *self,
   // one might think of using copyWithin but in SpiderMonkey it's implemented in JS!
   JS::RootedValue elementVal(GLOBAL_CX);
   for (Py_ssize_t inputIdx = 0; inputIdx < input_size; inputIdx++) {
-    JS_GetElement(GLOBAL_CX, self->jsObject, inputIdx, &elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, inputIdx, &elementVal);
     for (Py_ssize_t repeatIdx = 0; repeatIdx < n; repeatIdx++) {
       JS_SetElement(GLOBAL_CX, jCombinedArray, repeatIdx * input_size + inputIdx, elementVal);
     }
@@ -659,7 +673,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_repeat(JSArrayProxy *self,
   JS::RootedValue jCombinedArrayValue(GLOBAL_CX);
   jCombinedArrayValue.setObjectOrNull(jCombinedArray);
 
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject)), &jCombinedArrayValue)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), &jCombinedArrayValue)->getPyObject();
 }
 
 int JSArrayProxyMethodDefinitions::JSArrayProxy_contains(JSArrayProxy *self, PyObject *element) {
@@ -667,10 +681,10 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_contains(JSArrayProxy *self, PyO
   int cmp;
 
   Py_ssize_t numElements = JSArrayProxy_length(self);
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject));
+  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
   for (index = 0, cmp = 0; cmp == 0 && index < numElements; ++index) {
     JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsObject, index, elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
     PyObject *item = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
     Py_INCREF(item);
     cmp = PyObject_RichCompareBool(item, element, Py_EQ);
@@ -684,7 +698,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_inplace_concat(JSArrayProx
   Py_ssize_t valueLength = Py_SIZE(value);
 
   // allocate extra spacePy_SIZE
-  JS::SetArrayLength(GLOBAL_CX, self->jsObject, selfLength + valueLength);
+  JS::SetArrayLength(GLOBAL_CX, self->jsArray, selfLength + valueLength);
 
   JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, value));
   JS::RootedObject jRootedValue = JS::RootedObject(GLOBAL_CX, jValue.toObjectOrNull());
@@ -692,7 +706,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_inplace_concat(JSArrayProx
   JS::RootedValue elementVal(GLOBAL_CX);
   for (Py_ssize_t inputIdx = 0; inputIdx < valueLength; inputIdx++) {
     JS_GetElement(GLOBAL_CX, jRootedValue, inputIdx, &elementVal);
-    JS_SetElement(GLOBAL_CX, self->jsObject, selfLength + inputIdx, elementVal);
+    JS_SetElement(GLOBAL_CX, self->jsArray, selfLength + inputIdx, elementVal);
   }
 
   Py_INCREF(self);
@@ -716,15 +730,15 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_inplace_repeat(JSArrayProx
     return PyErr_NoMemory();
   }
 
-  JS::SetArrayLength(GLOBAL_CX, self->jsObject, input_size * n);
+  JS::SetArrayLength(GLOBAL_CX, self->jsArray, input_size * n);
 
   // repeat within self
   // one might think of using copyWithin but in SpiderMonkey it's implemented in JS!
   JS::RootedValue elementVal(GLOBAL_CX);
   for (Py_ssize_t inputIdx = 0; inputIdx < input_size; inputIdx++) {
-    JS_GetElement(GLOBAL_CX, self->jsObject, inputIdx, &elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, inputIdx, &elementVal);
     for (Py_ssize_t repeatIdx = 0; repeatIdx < n; repeatIdx++) {
-      JS_SetElement(GLOBAL_CX, self->jsObject, repeatIdx * input_size + inputIdx, elementVal);
+      JS_SetElement(GLOBAL_CX, self->jsArray, repeatIdx * input_size + inputIdx, elementVal);
     }
   }
 
@@ -733,7 +747,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_inplace_repeat(JSArrayProx
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_clear(JSArrayProxy *self) {
-  JS::SetArrayLength(GLOBAL_CX, self->jsObject, 0);
+  JS::SetArrayLength(GLOBAL_CX, self->jsArray, 0);
   Py_RETURN_NONE;
 }
 
@@ -743,10 +757,10 @@ int JSArrayProxyMethodDefinitions::JSArrayProxy_clear_slot(JSArrayProxy *self) {
 }
 
 int JSArrayProxyMethodDefinitions::JSArrayProxy_traverse(JSArrayProxy *self, visitproc visit, void *arg) {
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject));
+  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
   for (Py_ssize_t i = JSArrayProxy_length(self); --i >= 0; ) {
     JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsObject, i, elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, i, elementVal);
     Py_VISIT(pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject());
   }
   return 0;
@@ -757,19 +771,19 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_copy(JSArrayProxy *self) {
   jArgs[0].setInt32(0);
   jArgs[1].setInt32(JSArrayProxy_length(self));
   JS::RootedValue *jReturnedArray = new JS::RootedValue(GLOBAL_CX);
-  if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "slice", jArgs, jReturnedArray)) {
+  if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "slice", jArgs, jReturnedArray)) {
     PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
     return NULL;
   }
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject)), jReturnedArray)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), jReturnedArray)->getPyObject();
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_append(JSArrayProxy *self, PyObject *value) {
   Py_ssize_t len = JSArrayProxy_length(self);
 
-  JS::SetArrayLength(GLOBAL_CX, self->jsObject, len + 1);
+  JS::SetArrayLength(GLOBAL_CX, self->jsArray, len + 1);
   JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, value));
-  JS_SetElement(GLOBAL_CX, self->jsObject, len, jValue);
+  JS_SetElement(GLOBAL_CX, self->jsArray, len, jValue);
 
   Py_RETURN_NONE;
 }
@@ -816,7 +830,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_insert(JSArrayProxy *self,
   jArgs[2].set(jsTypeFactory(GLOBAL_CX, value));
 
   JS::RootedValue jReturnedArray(GLOBAL_CX);
-  if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "splice", jArgs, &jReturnedArray)) {
+  if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "splice", jArgs, &jReturnedArray)) {
     PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
     return NULL;
   }
@@ -839,14 +853,14 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_extend(JSArrayProxy *self,
 
     Py_ssize_t m = JSArrayProxy_length(self);
 
-    JS::SetArrayLength(GLOBAL_CX, self->jsObject, m + n);
+    JS::SetArrayLength(GLOBAL_CX, self->jsArray, m + n);
 
     // populate the end of self with iterable's items.
     PyObject **src = PySequence_Fast_ITEMS(iterable);
     for (Py_ssize_t i = 0; i < n; i++) {
       PyObject *o = src[i];
       JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, o));
-      JS_SetElement(GLOBAL_CX, self->jsObject, m + i, jValue);
+      JS_SetElement(GLOBAL_CX, self->jsArray, m + i, jValue);
     }
 
     Py_DECREF(iterable);
@@ -875,9 +889,9 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_extend(JSArrayProxy *self,
         break;
       }
 
-      JS::SetArrayLength(GLOBAL_CX, self->jsObject, len + 1);
+      JS::SetArrayLength(GLOBAL_CX, self->jsArray, len + 1);
       JS::RootedValue jValue(GLOBAL_CX, jsTypeFactory(GLOBAL_CX, item));
-      JS_SetElement(GLOBAL_CX, self->jsObject, len, jValue);
+      JS_SetElement(GLOBAL_CX, self->jsArray, len, jValue);
       len++;
     }
 
@@ -928,7 +942,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_pop(JSArrayProxy *self, Py
   jArgs[1].setInt32(1);
 
   JS::RootedValue jReturnedArray(GLOBAL_CX);
-  if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "splice", jArgs, &jReturnedArray)) {
+  if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "splice", jArgs, &jReturnedArray)) {
     PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
     return NULL;
   }
@@ -938,16 +952,16 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_pop(JSArrayProxy *self, Py
   JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
   JS_GetElement(GLOBAL_CX, rootedReturnedArray, 0, elementVal);
 
-  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject)), elementVal)->getPyObject();
+  return pyTypeFactory(GLOBAL_CX, new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray)), elementVal)->getPyObject();
 }
 
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_remove(JSArrayProxy *self, PyObject *value) {
   Py_ssize_t selfSize = JSArrayProxy_length(self);
 
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject));
+  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
   for (Py_ssize_t index = 0; index < selfSize; index++) {
     JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsObject, index, elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
     PyObject *obj = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
     Py_INCREF(obj);
     int cmp = PyObject_RichCompareBool(obj, value, Py_EQ);
@@ -957,7 +971,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_remove(JSArrayProxy *self,
       jArgs[0].setInt32(index);
       jArgs[1].setInt32(1);
       JS::RootedValue jReturnedArray(GLOBAL_CX);
-      if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "splice", jArgs, &jReturnedArray)) {
+      if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "splice", jArgs, &jReturnedArray)) {
         PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
         return NULL;
       }
@@ -1010,10 +1024,10 @@ skip_optional:
     }
   }
 
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject));
+  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
   for (Py_ssize_t index = start; index < stop && index < selfSize; index++) {
     JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsObject, index, elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
     PyObject *obj = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
     Py_INCREF(obj);
     int cmp = PyObject_RichCompareBool(obj, value, Py_EQ);
@@ -1034,10 +1048,10 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_count(JSArrayProxy *self,
   Py_ssize_t count = 0;
 
   Py_ssize_t length = JSArrayProxy_length(self);
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsObject));
+  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(self->jsArray));
   for (Py_ssize_t index = 0; index < length; index++) {
     JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX);
-    JS_GetElement(GLOBAL_CX, self->jsObject, index, elementVal);
+    JS_GetElement(GLOBAL_CX, self->jsArray, index, elementVal);
     PyObject *obj = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
     Py_INCREF(obj);
     int cmp = PyObject_RichCompareBool(obj, value, Py_EQ);
@@ -1055,7 +1069,7 @@ PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_count(JSArrayProxy *self,
 PyObject *JSArrayProxyMethodDefinitions::JSArrayProxy_reverse(JSArrayProxy *self) {
   if (JSArrayProxy_length(self) > 1) {
     JS::RootedValue jReturnedArray(GLOBAL_CX);
-    if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "reverse", JS::HandleValueArray::empty(), &jReturnedArray)) {
+    if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "reverse", JS::HandleValueArray::empty(), &jReturnedArray)) {
       PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
       return NULL;
     }
@@ -1085,17 +1099,17 @@ static bool sort_compare_key_func(JSContext *cx, unsigned argc, JS::Value *vp) {
   bool reverse = reverseValue.toBoolean();
 
 
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(&args.callee()));
+  JS::RootedObject *global = new JS::RootedObject(cx, JS::GetNonCCWObjectGlobal(&args.callee()));
 
-  JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX, args[0]);
-  PyObject *args_0 = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+  JS::RootedValue *elementVal = new JS::RootedValue(cx, args[0]);
+  PyObject *args_0 = pyTypeFactory(cx, global, elementVal)->getPyObject();
   PyObject *args_0_result = PyObject_CallFunction(keyfunc, "O", args_0);
   if (!args_0_result) {
     return false;
   }
 
-  elementVal = new JS::RootedValue(GLOBAL_CX, args[1]);
-  PyObject *args_1 = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+  elementVal = new JS::RootedValue(cx, args[1]);
+  PyObject *args_1 = pyTypeFactory(cx, global, elementVal)->getPyObject();
   PyObject *args_1_result = PyObject_CallFunction(keyfunc, "O", args_1);
   if (!args_1_result) {
     return false;
@@ -1135,13 +1149,13 @@ static bool sort_compare_default(JSContext *cx, unsigned argc, JS::Value *vp) {
   }
   bool reverse = reverseValue.toBoolean();
 
-  JS::RootedObject *global = new JS::RootedObject(GLOBAL_CX, JS::GetNonCCWObjectGlobal(&args.callee()));
+  JS::RootedObject *global = new JS::RootedObject(cx, JS::GetNonCCWObjectGlobal(&args.callee()));
 
-  JS::RootedValue *elementVal = new JS::RootedValue(GLOBAL_CX, args[0]);
-  PyObject *args_0 = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+  JS::RootedValue *elementVal = new JS::RootedValue(cx, args[0]);
+  PyObject *args_0 = pyTypeFactory(cx, global, elementVal)->getPyObject();
 
-  elementVal = new JS::RootedValue(GLOBAL_CX, args[1]);
-  PyObject *args_1 = pyTypeFactory(GLOBAL_CX, global, elementVal)->getPyObject();
+  elementVal = new JS::RootedValue(cx, args[1]);
+  PyObject *args_1 = pyTypeFactory(cx, global, elementVal)->getPyObject();
 
   int cmp = PyObject_RichCompareBool(args_0, args_1, Py_LT);
   if (cmp > 0) {
@@ -1245,18 +1259,29 @@ skip_optional_kwonly:
 
           JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
           jArgs[0].setObject(*funObj);
-          if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "sort", jArgs, &jReturnedArray)) {
+          if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
             if (!PyErr_Occurred()) {
               PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
             }
             return NULL;
           }
+
+          // cleanup
+          if (!JS_DeleteProperty(GLOBAL_CX, funObj, "_key_func_param")) {
+            PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
+            return NULL;
+          }
+
+          if (!JS_DeleteProperty(GLOBAL_CX, funObj, "_reverse_param")) {
+            PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
+            return NULL;
+          }
         }
         else {
           // two-arg js-style
           JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
           jArgs[0].set(jsTypeFactory(GLOBAL_CX, keyfunc));
-          if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "sort", jArgs, &jReturnedArray)) {
+          if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
             PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
             return NULL;
           }
@@ -1274,7 +1299,7 @@ skip_optional_kwonly:
           // we got a JS compare function, use it as-is
           JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
           jArgs[0].set(jsTypeFactory(GLOBAL_CX, keyfunc));
-          if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "sort", jArgs, &jReturnedArray)) {
+          if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
             PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
             return NULL;
           }
@@ -1302,7 +1327,7 @@ skip_optional_kwonly:
 
           JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
           jArgs[0].setObject(*funObj);
-          if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "sort", jArgs, &jReturnedArray)) {
+          if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
             if (!PyErr_Occurred()) {
               PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
             }
@@ -1329,7 +1354,7 @@ skip_optional_kwonly:
 
       JS::Rooted<JS::ValueArray<1>> jArgs(GLOBAL_CX);
       jArgs[0].setObject(*funObj);
-      if (!JS_CallFunctionName(GLOBAL_CX, self->jsObject, "sort", jArgs, &jReturnedArray)) {
+      if (!JS_CallFunctionName(GLOBAL_CX, self->jsArray, "sort", jArgs, &jReturnedArray)) {
         PyErr_Format(PyExc_SystemError, "%s JSAPI call failed", JSArrayProxyType.tp_name);
         return NULL;
       }
