@@ -16,7 +16,8 @@
 #include "include/FuncType.hh"
 #include "include/JSObjectProxy.hh"
 #include "include/JSArrayProxy.hh"
-#include "include/PyProxyHandler.hh"
+#include "include/PyDictProxyHandler.hh"
+#include "include/PyListProxyHandler.hh"
 #include "include/pyTypeFactory.hh"
 #include "include/StrType.hh"
 #include "include/IntType.hh"
@@ -165,16 +166,22 @@ JS::Value jsTypeFactory(JSContext *cx, PyObject *object) {
     returnType.setObject(*((JSObjectProxy *)object)->jsObject);
   }
   else if (PyObject_TypeCheck(object, &JSArrayProxyType)) {
-    returnType.setObject(*((JSArrayProxy *)object)->jsObject);
+    returnType.setObject(*((JSArrayProxy *)object)->jsArray);
   }
   else if (PyDict_Check(object) || PyList_Check(object)) {
     JS::RootedValue v(cx);
     JSObject *proxy;
     if (PyList_Check(object)) {
-      proxy = js::NewProxyObject(cx, new PyListProxyHandler(object), v, NULL);
+      JS::RootedObject arrayPrototype(cx);
+      JS_GetClassPrototype(cx, JSProto_Array, &arrayPrototype); // so that instanceof will work, not that prototype methods will
+      proxy = js::NewProxyObject(cx, new PyListProxyHandler(object), v, arrayPrototype.get());
     } else {
-      proxy = js::NewProxyObject(cx, new PyProxyHandler(object), v, NULL);
+      JS::RootedObject objectPrototype(cx);
+      JS_GetClassPrototype(cx, JSProto_Object, &objectPrototype); // so that instanceof will work, not that prototype methods will
+      proxy = js::NewProxyObject(cx, new PyDictProxyHandler(object), v, objectPrototype.get());
     }
+    Py_INCREF(object); // TODO leak! clean up in finalize
+    JS::SetReservedSlot(proxy, PyObjectSlot, JS::PrivateValue(object));
     returnType.setObject(*proxy);
   }
   else if (object == Py_None) {
@@ -239,7 +246,9 @@ bool callPyFunc(JSContext *cx, unsigned int argc, JS::Value *vp) {
   JS::RootedObject *thisv = new JS::RootedObject(cx);
   JS_ValueToObject(cx, callargs.thisv(), thisv);
 
-  if (!callargs.length()) {
+  unsigned int callArgsLength = callargs.length();
+
+  if (!callArgsLength) {
     #if PY_VERSION_HEX >= 0x03090000
     PyObject *pyRval = PyObject_CallNoArgs(pyFunc);
     #else
@@ -255,8 +264,8 @@ bool callPyFunc(JSContext *cx, unsigned int argc, JS::Value *vp) {
   }
 
   // populate python args tuple
-  PyObject *pyArgs = PyTuple_New(callargs.length());
-  for (size_t i = 0; i < callargs.length(); i++) {
+  PyObject *pyArgs = PyTuple_New(callArgsLength);
+  for (size_t i = 0; i < callArgsLength; i++) {
     JS::RootedValue *jsArg = new JS::RootedValue(cx, callargs[i]);
     PyType *pyArg = pyTypeFactory(cx, thisv, jsArg);
     if (!pyArg) return false; // error occurred
