@@ -17,7 +17,12 @@
 #include "include/DateType.hh"
 #include "include/FloatType.hh"
 #include "include/FuncType.hh"
+#include "include/JSArrayIterProxy.hh"
 #include "include/JSArrayProxy.hh"
+#include "include/JSObjectIterProxy.hh"
+#include "include/JSObjectKeysProxy.hh"
+#include "include/JSObjectValuesProxy.hh"
+#include "include/JSObjectItemsProxy.hh"
 #include "include/JSObjectProxy.hh"
 #include "include/PyType.hh"
 #include "include/pyTypeFactory.hh"
@@ -79,6 +84,7 @@ PyTypeObject JSObjectProxyType = {
   .tp_as_number = &JSObjectProxy_number_methods,
   .tp_as_sequence = &JSObjectProxy_sequence_methods,
   .tp_as_mapping = &JSObjectProxy_mapping_methods,
+  .tp_hash = PyObject_HashNotImplemented,
   .tp_getattro = (getattrofunc)JSObjectProxyMethodDefinitions::JSObjectProxy_get,
   .tp_setattro = (setattrofunc)JSObjectProxyMethodDefinitions::JSObjectProxy_assign,
   .tp_flags = Py_TPFLAGS_DEFAULT
@@ -114,6 +120,96 @@ PyTypeObject JSArrayProxyType = {
   .tp_new = JSArrayProxyMethodDefinitions::JSArrayProxy_new,
 };
 
+PyTypeObject JSArrayIterProxyType = {
+  .ob_base = PyVarObject_HEAD_INIT(NULL, 0)
+  .tp_name = "pythonmonkey.JSArrayIterProxy",
+  .tp_basicsize = sizeof(JSArrayIterProxy),
+  .tp_itemsize = 0,
+  .tp_dealloc = (destructor)JSArrayIterProxyMethodDefinitions::JSArrayIterProxy_dealloc,
+  .tp_getattro = PyObject_GenericGetAttr,
+  .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
+  .tp_doc = PyDoc_STR("Javascript Array proxy iterator"),
+  .tp_traverse =  (traverseproc)JSArrayIterProxyMethodDefinitions::JSArrayIterProxy_traverse,
+  .tp_iter = (getiterfunc)JSArrayIterProxyMethodDefinitions::JSArrayIterProxy_iter,
+  .tp_iternext = (iternextfunc)JSArrayIterProxyMethodDefinitions::JSArrayIterProxy_next,
+  .tp_methods = JSArrayIterProxy_methods,
+  .tp_base = &PyListIter_Type
+};
+
+PyTypeObject JSObjectIterProxyType = {
+  .ob_base = PyVarObject_HEAD_INIT(NULL, 0)
+  .tp_name = "pythonmonkey.JSObjectIterProxy",
+  .tp_basicsize = sizeof(JSObjectIterProxy),
+  .tp_itemsize = 0,
+  .tp_dealloc = (destructor)JSObjectIterProxyMethodDefinitions::JSObjectIterProxy_dealloc,
+  .tp_getattro = PyObject_GenericGetAttr,
+  .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
+  .tp_doc = PyDoc_STR("Javascript Object proxy key iterator"),
+  .tp_traverse =  (traverseproc)JSObjectIterProxyMethodDefinitions::JSObjectIterProxy_traverse,
+  .tp_iter = (getiterfunc)JSObjectIterProxyMethodDefinitions::JSObjectIterProxy_iter,
+  .tp_iternext = (iternextfunc)JSObjectIterProxyMethodDefinitions::JSObjectIterProxy_nextkey,
+  .tp_methods = JSObjectIterProxy_methods,
+  .tp_base = &PyDictIterKey_Type
+};
+
+PyTypeObject JSObjectKeysProxyType = {
+  .ob_base = PyVarObject_HEAD_INIT(NULL, 0)
+  .tp_name = "pythonmonkey.JSObjectKeysProxy",
+  .tp_basicsize = sizeof(JSObjectKeysProxy),
+  .tp_itemsize = 0,
+  .tp_dealloc = (destructor)JSObjectKeysProxyMethodDefinitions::JSObjectKeysProxy_dealloc,
+  .tp_repr = (reprfunc)JSObjectKeysProxyMethodDefinitions::JSObjectKeysProxy_repr,
+  .tp_as_number = &JSObjectKeysProxy_number_methods,
+  .tp_as_sequence = &JSObjectKeysProxy_sequence_methods,
+  .tp_getattro = PyObject_GenericGetAttr,
+  .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
+  .tp_doc = PyDoc_STR("Javascript Object Keys proxy"),
+  .tp_traverse =  (traverseproc)JSObjectKeysProxyMethodDefinitions::JSObjectKeysProxy_traverse,
+  .tp_richcompare = (richcmpfunc)JSObjectKeysProxyMethodDefinitions::JSObjectKeysProxy_richcompare,
+  .tp_iter = (getiterfunc)JSObjectKeysProxyMethodDefinitions::JSObjectKeysProxy_iter,
+  .tp_methods = JSObjectKeysProxy_methods,
+  .tp_getset = JSObjectKeysProxy_getset,
+  .tp_base = &PyDictKeys_Type
+};
+
+PyTypeObject JSObjectValuesProxyType = {
+  .ob_base = PyVarObject_HEAD_INIT(NULL, 0)
+  .tp_name = "pythonmonkey.JSObjectValuesProxy",
+  .tp_basicsize = sizeof(JSObjectValuesProxy),
+  .tp_itemsize = 0,
+  .tp_dealloc = (destructor)JSObjectValuesProxyMethodDefinitions::JSObjectValuesProxy_dealloc,
+  .tp_repr = (reprfunc)JSObjectValuesProxyMethodDefinitions::JSObjectValuesProxy_repr,
+  .tp_as_sequence = &JSObjectValuesProxy_sequence_methods,
+  .tp_getattro = PyObject_GenericGetAttr,
+  .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
+  .tp_doc = PyDoc_STR("Javascript Object Values proxy"),
+  .tp_traverse =  (traverseproc)JSObjectValuesProxyMethodDefinitions::JSObjectValuesProxy_traverse,
+  .tp_iter = (getiterfunc)JSObjectValuesProxyMethodDefinitions::JSObjectValuesProxy_iter,
+  .tp_methods = JSObjectValuesProxy_methods,
+  .tp_getset = JSObjectValuesProxy_getset,
+  .tp_base = &PyDictValues_Type
+};
+
+PyTypeObject JSObjectItemsProxyType = {
+  .ob_base = PyVarObject_HEAD_INIT(NULL, 0)
+  .tp_name = "pythonmonkey.JSObjectItemsProxy",
+  .tp_basicsize = sizeof(JSObjectItemsProxy),
+  .tp_itemsize = 0,
+  .tp_dealloc = (destructor)JSObjectItemsProxyMethodDefinitions::JSObjectItemsProxy_dealloc,
+  .tp_repr = (reprfunc)JSObjectItemsProxyMethodDefinitions::JSObjectItemsProxy_repr,
+  // .tp_as_number = defaults are fine
+  .tp_as_sequence = &JSObjectItemsProxy_sequence_methods,
+  .tp_getattro = PyObject_GenericGetAttr,
+  .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
+  .tp_doc = PyDoc_STR("Javascript Object Items proxy"),
+  .tp_traverse =  (traverseproc)JSObjectItemsProxyMethodDefinitions::JSObjectItemsProxy_traverse,
+  // .tp_richcompare = TODO tuple support
+  .tp_iter = (getiterfunc)JSObjectItemsProxyMethodDefinitions::JSObjectItemsProxy_iter,
+  .tp_methods = JSObjectItemsProxy_methods,
+  .tp_getset = JSObjectItemsProxy_getset,
+  .tp_base = &PyDictKeys_Type
+};
+
 static void cleanup() {
   delete autoRealm;
   delete global;
@@ -381,7 +477,10 @@ PyMODINIT_FUNC PyInit_pythonmonkey(void)
     return NULL;
   }
 
-  JS::RealmOptions options;
+  JS::RealmCreationOptions creationOptions = JS::RealmCreationOptions();
+  JS::RealmBehaviors behaviours = JS::RealmBehaviors();
+  creationOptions.setIteratorHelpersEnabled(true);
+  JS::RealmOptions options = JS::RealmOptions(creationOptions, behaviours);
   static JSClass globalClass = {"global", JSCLASS_GLOBAL_FLAGS, &JS::DefaultGlobalClassOps};
   global = new JS::RootedObject(GLOBAL_CX, JS_NewGlobalObject(GLOBAL_CX, &globalClass, nullptr, JS::FireOnNewGlobalHook, options));
   if (!global) {
@@ -418,6 +517,16 @@ PyMODINIT_FUNC PyInit_pythonmonkey(void)
     return NULL;
   if (PyType_Ready(&JSArrayProxyType) < 0)
     return NULL;
+  if (PyType_Ready(&JSArrayIterProxyType) < 0)
+    return NULL;
+  if (PyType_Ready(&JSObjectIterProxyType) < 0)
+    return NULL;
+  if (PyType_Ready(&JSObjectKeysProxyType) < 0)
+    return NULL;
+  if (PyType_Ready(&JSObjectValuesProxyType) < 0)
+    return NULL;
+  if (PyType_Ready(&JSObjectItemsProxyType) < 0)
+    return NULL;
 
   pyModule = PyModule_Create(&pythonmonkey);
   if (pyModule == NULL)
@@ -450,6 +559,41 @@ PyMODINIT_FUNC PyInit_pythonmonkey(void)
     return NULL;
   }
 
+  Py_INCREF(&JSArrayIterProxyType);
+  if (PyModule_AddObject(pyModule, "JSArrayIterProxy", (PyObject *)&JSArrayIterProxyType) < 0) {
+    Py_DECREF(&JSArrayIterProxyType);
+    Py_DECREF(pyModule);
+    return NULL;
+  }
+
+  Py_INCREF(&JSObjectIterProxyType);
+  if (PyModule_AddObject(pyModule, "JSObjectIterProxy", (PyObject *)&JSObjectIterProxyType) < 0) {
+    Py_DECREF(&JSObjectIterProxyType);
+    Py_DECREF(pyModule);
+    return NULL;
+  }
+
+  Py_INCREF(&JSObjectKeysProxyType);
+  if (PyModule_AddObject(pyModule, "JSObjectKeysProxy", (PyObject *)&JSObjectKeysProxyType) < 0) {
+    Py_DECREF(&JSObjectKeysProxyType);
+    Py_DECREF(pyModule);
+    return NULL;
+  }
+
+  Py_INCREF(&JSObjectValuesProxyType);
+  if (PyModule_AddObject(pyModule, "JSObjectValuesProxy", (PyObject *)&JSObjectValuesProxyType) < 0) {
+    Py_DECREF(&JSObjectValuesProxyType);
+    Py_DECREF(pyModule);
+    return NULL;
+  }
+
+  Py_INCREF(&JSObjectItemsProxyType);
+  if (PyModule_AddObject(pyModule, "JSObjectItemsProxy", (PyObject *)&JSObjectItemsProxyType) < 0) {
+    Py_DECREF(&JSObjectItemsProxyType);
+    Py_DECREF(pyModule);
+    return NULL;
+  }
+
   if (PyModule_AddObject(pyModule, "SpiderMonkeyError", SpiderMonkeyError)) {
     Py_DECREF(pyModule);
     return NULL;
